#!/usr/bin/env python
####################################################################################################
# NAME
#    <NAME> - test utility functions
#
# SYNOPSIS
#    <NAME>
#
# AUTHOR
#    Written by Florian Barras (florian@barras.io).
#
# COPYRIGHT
#    Copyright © 2013-2022 Florian Barras <https://barras.io>.
#    The MIT License (MIT) <https://opensource.org/licenses/MIT>.
####################################################################################################

import timeit
import unittest

from nutil.ts import *

####################################################################################################
# TEST CONSTANTS
####################################################################################################

_TEST_CONSTANTS___________________________________ = ''

PRECISION = 14  # decimals

TEST_COUNT = 10

####################################################################################################
# TEST CLASSES
####################################################################################################

__TEST_CLASSES____________________________________ = ''


class Test(unittest.TestCase):

	def assert_equals(self, first, second, precision=PRECISION):
		if is_collection(first):
			for i in range(len(first)):
				self.assert_equals(first[i], second[i], precision=precision)
		else:
			if is_number(first) and is_number(second):
				self.assertAlmostEqual(first, second, places=precision)
			else:
				self.assertEqual(first, second)


class TestCommon(Test):

	def test(self):
		# Initialize the collections
		d = to_dict(reverse(range(100)))
		s = to_series(d, name='a')
		df = concat_cols(s, to_frame(reverse(get_values(s)), names='b'))
		g0 = df.groupby(by=get_index(df), axis=0)
		g1 = df.groupby(by={k: 'group' for k in get_keys(df)}, axis=1)
		f = np.sum

		test('Test the dictionary functions')
		self.get_items(d)

		self.apply(d, f)
		self.apply(d.copy(), f, inplace=True)

		test('Test the series functions')
		self.get_items(s)

		self.get_rows(s)
		self.get_cols(s)

		self.apply(s, f)
		self.apply(s, f, axis=0)
		self.apply(s.copy(), f, inplace=True)

		test('Test the frame functions')
		self.get_items(df)

		self.get_rows(df)
		self.get_cols(df)

		self.apply(df, f)
		self.apply(df, f, axis=0)
		self.apply(df, f, axis=1)
		self.apply(df.copy(), f, inplace=True)

		test('Test the group functions')
		self.get_items(g0)
		self.get_items(g1)

		self.apply(g0, f)
		self.apply(g1, f)

	def apply(self, c, f, axis=None, inplace=False):
		t = timeit.timeit(stmt=lambda: apply(c, f, axis=axis, inplace=inplace), number=TEST_COUNT)
		test('Applied', f.__name__, 'on', count(c, axis=None), 'items', TEST_COUNT, 'times in',
		     round(t, 6), '[s]')

	def get_items(self, c):
		t = timeit.timeit(stmt=lambda: get_items(c, inclusion=range(len(c)),
		                                         exclusion=range(int(len(c) / 2))),
		                  number=TEST_COUNT)
		test(len(c), 'items retrieved', TEST_COUNT, 'times in', round(t, 6), '[s]')

	def get_rows(self, c):
		t = timeit.timeit(stmt=lambda: get_rows(c), number=TEST_COUNT)
		test(count_rows(c), 'rows retrieved', TEST_COUNT, 'times in', round(t, 6), '[s]')

	def get_cols(self, c):
		t = timeit.timeit(stmt=lambda: get_cols(c), number=TEST_COUNT)
		test(count_cols(c), 'cols retrieved', TEST_COUNT, 'times in', round(t, 6), '[s]')


class TestTimeSeries(Test):

	def test(self):
		date_to = get_datetime()
		date_from = date_to - 2 * YEAR
		index = create_datetime_sequence(date_from, date_to)
		s = to_series(range(len(index)), index=index)

		test('Test the time series functions')
		s = transform_series(s, freq=Frequency.MONTHS, group=Group.LAST,
		                     transformation=Transformation.LOG_RETURNS)
		test(get_first(get_index(s)), '=', get_next_month_end(date_from))
		self.assert_equals(to_stamp(get_first(get_index(s))),
		                   to_stamp(get_next_month_end(date_from)))
		test(find_nearest_freq(s), '=', Frequency.MONTHS)
		self.assert_equals(find_nearest_freq(s).value, Frequency.MONTHS.value)
		test(find_nearest_group(s), '=', Group.LAST)
		self.assert_equals(find_nearest_group(s).value, Group.LAST.value)


####################################################################################################
# TEST MAIN
####################################################################################################

__TEST_MAIN_______________________________________ = ''


def main():
	"""Tests the library."""
	unittest.main()


if __name__ == '__main__':
	main()
